//----------------------------------------------------------------------
//
// Module Name:  amtu-ppc.c
//
// Include File:  none
//
// Description:   Code for Abstract Machine Test sparc64 privilege test.
//
// Notes:  This module performs the machine specific privilege tests
//              to ensure that the underlying hardware is still enforcing
//              the appropriate control mechanisms.
// -----------------------------------------------------------------
// LANGUAGE:     C
//
// (C) Oracle Inc, 2014
// Licensed under the Common Public License v. 1.0
// -----------------------------------------------------------------
//
// Change Activity:
// DATE     PGMR      COMMENTS
// -------- --------- ----------------------
// 05/03/14 J.Marchesi Created
//----------------------------------------------------------------------

#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <syscall.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <syslog.h>
#include "amtu.h"

#if defined(HAVE_SPARC64)

/************************************************************************/
/*                                                                      */
/* FUNCTION: catchfault                                                 */
/*                                                                      */
/* PURPOSE: Signal handler to catch the segmentation violation which is */
/*          expected when trying to execute privileged instructions     */
/*          without privilege.                                          */
/*                                                                      */
/************************************************************************/
void catchfault(int sig)
{
	if (debug) {
		printf("caught the fault %d\n", sig);
	}
	exit(0);
}

/************************************************************************/
/*                                                                      */
/* MACRO: MY_LOG                                                        */
/*                                                                      */
/* PURPOSE: Log a text using whatever appropriate logging facility.     */
/*          Used in TEST_INSN and amtu_priv.                            */
/*                                                                      */
/************************************************************************/
#ifdef HAVE_LIBLAUS
# define MY_LOG(str,level)  LAUS_LOG((str))
#else
# define MY_LOG(str,level)  AUDIT_LOG(str, level)
#endif

/************************************************************************/
/*                                                                      */
/* MACRO: TEST_INSN                                                     */
/*                                                                      */
/* PURPOSE: Execute a single instruction in a child process and make    */
/*          sure it segfaults when executing a privileged instruction.  */
/*          Used in amtu_priv.                                          */
/*                                                                      */
/************************************************************************/
#define TEST_INSN(INSN,ARGS)						\
	do {								\
	  pid_t pid, wpid;						\
	  int stat;							\
	  pid = fork();							\
	  if (pid == 0) {						\
	    /* child */							\
	    if (debug)							\
	      printf("%s test: ", INSN);				\
	    __asm__ (INSN " " ARGS);					\
	    exit(-1);							\
	  } else if (pid == -1) {					\
	    /* error condition */					\
	    fprintf(stderr, "Privilege Separation Test FAILED (%s)!\n", INSN); \
	    MY_LOG("amtu failed privilege separation on " INSN, 0);	\
	    exit (-1);							\
	  }								\
	  /* parent */							\
	  wpid = wait(&stat);						\
	  if (!(WIFEXITED(stat) && (WEXITSTATUS(stat) == 0))) {		\
	    fprintf(stderr, "Privilege Separation Test FAILED on %s!\n", INSN);	\
	    MY_LOG("amtu failed privilege separation on " INSN, 0);	\
	    return(-1);							\
	  }								\
	} while (0)

/************************************************************************/
/*                                                                      */
/* FUNCTION: amtu_priv                                                  */
/*                                                                      */
/* PURPOSE: Execute privileged instructions to ensure that they cannot  */
/*          legitimately be run in user mode.                            */
/*                                                                      */
/************************************************************************/
int amtu_priv(int argc, char *argv[])
{
	struct sigaction sig;

	printf("Executing Supervisor Mode Instructions Test...\n");

	sig.sa_handler = catchfault;
	sig.sa_flags = 0;
	sigemptyset(&sig.sa_mask);
	sigaction(SIGSEGV, &sig, NULL);
	sigaction(SIGILL, &sig, NULL);

	/* The following instructions are privileged in all
	   circumstances.  */

	TEST_INSN ("ALLCLEAN","");
	TEST_INSN ("INVALW",  "");
	TEST_INSN ("NORMALW",  "");
	TEST_INSN ("OTHERW",   "");
	TEST_INSN ("RESTORED", "");
	TEST_INSN ("RETRY", "");
	TEST_INSN ("SAVED", "");
#if 0
	/* These instructions require v9b or later on the
	   assembler, i.e ultrasparc3 or later.  */
	TEST_INSN ("RD", "%softint, %g0");
	TEST_INSN ("RD", "%stick_cmpr, %g0");
	TEST_INSN ("RD", "%cfr, %g0");
	TEST_INSN ("WR", "%g0, %g0, %softint_set");
	TEST_INSN ("WR", "%g0, %g0, %softint_clr");
	TEST_INSN ("WR", "%g0, %g0, %softint");
	TEST_INSN ("WR", "%g0, %g0, %stick_cmpr");
#endif
	TEST_INSN ("RDPR", "%tpc,        %g0");
	TEST_INSN ("RDPR", "%tnpc,       %g0");
	TEST_INSN ("RDPR", "%tstate,     %g0");
	TEST_INSN ("RDPR", "%tt,         %g0");
	TEST_INSN ("RDPR", "%tick,       %g0");
	TEST_INSN ("RDPR", "%tba,        %g0");
	TEST_INSN ("RDPR", "%pstate,     %g0");
	TEST_INSN ("RDPR", "%tl,         %g0");
	TEST_INSN ("RDPR", "%pil,        %g0");
	TEST_INSN ("RDPR", "%cwp,        %g0");
	TEST_INSN ("RDPR", "%cansave,    %g0");
	TEST_INSN ("RDPR", "%canrestore, %g0");
	TEST_INSN ("RDPR", "%cleanwin,   %g0");
	TEST_INSN ("RDPR", "%otherwin,   %g0");
	TEST_INSN ("RDPR", "%wstate,     %g0");
	TEST_INSN ("RDPR", "%gl,         %g0");
	TEST_INSN ("WRPR", "%g0, %g0, %tpc");
	TEST_INSN ("WRPR", "%g0, %g0, %tnpc");
	TEST_INSN ("WRPR", "%g0, %g0, %tstate");
	TEST_INSN ("WRPR", "%g0, %g0, %tt");
	TEST_INSN ("WRPR", "%g0, %g0, %tick");
	TEST_INSN ("WRPR", "%g0, %g0, %tba");
	TEST_INSN ("WRPR", "%g0, %g0, %pstate");
	TEST_INSN ("WRPR", " %g0, %g0, %tl");
	TEST_INSN ("WRPR", "%g0, %g0, %pil");
	TEST_INSN ("WRPR", "%g0, %g0, %cwp");
	TEST_INSN ("WRPR", "%g0, %g0, %cansave");
	TEST_INSN ("WRPR", "%g0, %g0, %canrestore");
	TEST_INSN ("WRPR", "%g0, %g0, %cleanwin");
	TEST_INSN ("WRPR", "%g0, %g0, %otherwin");
	TEST_INSN ("WRPR", "%g0, %g0, %wstate" );
	TEST_INSN ("WRPR", "%g0, %g0, %gl");

	/* The following instructions are privileged if the referred
	   ASI is privileted, i.e. the bit 7 of the ASI is 0.  */

	TEST_INSN ("CASA", "[%g0] #ASI_AS_IF_USER_PRIMARY,  %g0, %g0");
	TEST_INSN ("CASXA", "[%g0] #ASI_AS_IF_USER_PRIMARY, %g0, %g0");
	TEST_INSN ("LDSBA", "[%g0] #ASI_AS_IF_USER_PRIMARY, %g0");
	TEST_INSN ("LDSHA", "[%g0] #ASI_AS_IF_USER_PRIMARY, %g0");
	TEST_INSN ("LDSWA", "[%g0] #ASI_AS_IF_USER_PRIMARY, %g0");
	TEST_INSN ("LDUBA", "[%g0] #ASI_AS_IF_USER_PRIMARY, %g0");
	TEST_INSN ("LDUHA", "[%g0] #ASI_AS_IF_USER_PRIMARY, %g0");
	TEST_INSN ("LDUWA", "[%g0] #ASI_AS_IF_USER_PRIMARY, %g0");
	TEST_INSN ("LDXA",  "[%g0] #ASI_AS_IF_USER_PRIMARY, %g0");
	TEST_INSN ("STBA",  "%g0, [%g0] #ASI_AS_IF_USER_PRIMARY");
	TEST_INSN ("STHA",  "%g0, [%g0] #ASI_AS_IF_USER_PRIMARY");
	TEST_INSN ("STWA",  "%g0, [%g0] #ASI_AS_IF_USER_PRIMARY");
	TEST_INSN ("STXA",  "%g0, [%g0] #ASI_AS_IF_USER_PRIMARY");
	TEST_INSN ("LDA",   "[%g0] #ASI_AS_IF_USER_PRIMARY, %f0");
	TEST_INSN ("LDDA",  "[%g0] #ASI_AS_IF_USER_PRIMARY, %f0");
	TEST_INSN ("LDQA",  "[%g0] #ASI_AS_IF_USER_PRIMARY, %f0");
	TEST_INSN ("STA",   "%f0, [%g0] #ASI_AS_IF_USER_PRIMARY");
	TEST_INSN ("STDA",  "%f0, [%g0] #ASI_AS_IF_USER_PRIMARY");
	TEST_INSN ("STQA",  "%f0, [%g0] #ASI_AS_IF_USER_PRIMARY");

	/* The following instructions are hyperprivileged.  */

	TEST_INSN ("SIR",   "0x0");
#if 0
	/* These instructions require v9b or later on the
	   assembler, i.e ultrasparc3 or later.  */
	TEST_INSN ("WR",    "%g0, %g0, %stick");
	TEST_INSN ("WR",    "%g0, %g0, %cfr");
#endif
	TEST_INSN ("rdhpr", "%hpstate,       %g0");
	TEST_INSN ("rdhpr", "%htstate,       %g0");
	TEST_INSN ("rdhpr", "%hintp,         %g0");
	TEST_INSN ("rdhpr", "%htba,          %g0");
	TEST_INSN ("rdhpr", "%hver,          %g0");
	TEST_INSN ("rdhpr", "%hstick_cmpr,   %g0");
	TEST_INSN ("wrhpr", "%g0, %g0, %hpstate");
	TEST_INSN ("wrhpr", "%g0, %g0, %htstate");
	TEST_INSN ("wrhpr", "%g0, %g0, %hintp");
	TEST_INSN ("wrhpr", "%g0, %g0, %htba");
	TEST_INSN ("wrhpr", "%g0, %g0, %hstick_cmpr");
#if 0
	/* Not supported in binutils atm.  */
	TEST_INSN ("JPRIV", ".");
	TEST_INSN ("wrhpr", "%g0, %g0, %hstick_offset");
	TEST_INSN ("wrhpr", "%g0, %g0, %hstick_enable");
	TEST_INSN ("rdhpr", "%hstick_offset, %g0");
	TEST_INSN ("rdhpr", "%hstick_enable, %g0");
#endif


        MY_LOG ("amtu - Privileged Instruction Test succeeded", 1);
	printf("Privileged Instruction Test SUCCESS!\n");
	return(0);
}

#endif /* HAVE_SPARC64 */
